from enum import IntEnum
from typing import (
    Union,
    TypeVar,
    Any,
    Tuple,
    Dict,
    Optional,
    KeysView,
    ItemsView,
    ValuesView,
    Iterator,
)

from ._bonsai import ldapentry
from .errors import InvalidDN
from .ldapdn import LDAPDN
from .ldapvaluelist import LDAPValueList

MYPY = False

if MYPY:
    from .ldapconnection import BaseLDAPConnection

CT = TypeVar("CT", bound="BaseLDAPConnection")


class LDAPModOp(IntEnum):
    """ Enumeration for LDAP modification operations. """

    ADD = 0  #: For adding new values to the attribute.
    DELETE = 1  #: For deleting existing values from the attribute list.
    REPLACE = 2  #: For replacing the existing attribute values.


class LDAPEntry(ldapentry):
    def __init__(self, dn: Union[LDAPDN, str], conn: Optional[CT] = None) -> None:
        try:
            super().__init__(dn, conn)
            self.__extended_dn = None  # type: Optional[str]
        except InvalidDN as exc:
            # InvalidDN error caused by extended DN control.
            splitted_dn = str(dn).split(";")
            try:
                super().__init__(splitted_dn[-1], conn)
            except InvalidDN:
                raise exc from None
            self.__extended_dn = str(dn)

    def delete(self, timeout: Optional[float] = None, recursive: bool = False) -> Any:
        """
        Remove LDAP entry from the dictionary server.

        :param float timeout: time limit in seconds for the operation.
        :param bool recursive: remove every entry of the given subtree \
        recursively.
        :return: True, if the operation is finished.
        :rtype: bool
        """
        res = self.connection.delete(self.dn, timeout, recursive)
        for value in self.values():
            try:
                value.status = 2
            except AttributeError:
                continue  # Skip DN.
        return res

    def modify(self, timeout: Optional[float] = None) -> Any:
        """
        Send entry's modifications to the dictionary server.

        :param float timeout: time limit in seconds for the operation.
        :return: True, if the operation is finished.
        :rtype: bool
        """
        return self.connection._evaluate(super().modify(), timeout)

    def rename(
        self,
        newdn: Union[str, LDAPDN],
        timeout: Optional[float] = None,
        delete_old_rdn: bool = True,
    ) -> Any:
        """
        Change the entry's distinguished name.

        :param str|LDAPDN newdn: the new DN of the entry.
        :param float timeout: time limit in seconds for the operation.
        :param bool delete_old_rdn: remove old RDN with renaming.
        :return: True, if the operation is finished.
        :rtype: bool
        """
        if type(newdn) == LDAPDN:
            newdn = str(newdn)
        return self.connection._evaluate(super().rename(newdn, delete_old_rdn), timeout)

    def update(self, *args: Tuple, **kwds: Dict[str, Any]) -> None:
        """
        Update the LDAPEntry with the key/value pairs from other, overwriting existing keys.
        (Same as dict's update method.)
        """
        if args:
            if hasattr(args[0], "keys"):
                # Working with a dict on the parameter list.
                for key, value in args[0].items():
                    self.__setitem__(key, value)
            else:
                # Working with a sequence.
                for tup in args[0]:
                    if len(tup) != 2:
                        raise ValueError("Sequence element has more then 2 element.")
                    self.__setitem__(tup[0], tup[1])
        if kwds:
            # Key/value pairs are listed on the parameter list.
            for key, value in kwds.items():
                self.__setitem__(key, value)

    def clear(self) -> None:
        """ Remove all items from the dictionary. """
        keys = list(self.keys())
        for key in keys:
            try:
                del self[key]
            except TypeError:
                continue  # Skip DN.

    def get(self, key: str, default: Any = None) -> Any:
        """
        Return the value for `key` if `key` is in the LDAPEntry,
        else `default`. (Same as dict's get method.)
        """
        try:
            return self[key]
        except KeyError:
            return default

    def pop(self, *args: Tuple) -> Any:
        """
        LDAPEntry.pop(k[,d]) -> v, remove specified key and return the
        corresponding value. If key is not found, d is returned if given,
        otherwise KeyError is raised.

        :param key: the key.
        :param dflt: if key is not found, d is returned.
        :return: the value from the LDAPEntry.
        """
        if len(args) > 2:
            raise TypeError("pop expected at most 2 arguments, got %d" % len(args))
        try:
            key = args[0]
            value = self[key]
            del self[key]
            return value
        except IndexError:
            raise TypeError("pop expected at least 1 arguments, got 0")
        except KeyError as err:
            try:
                dflt = args[1]
                return dflt
            except IndexError:
                raise err

    def popitem(self) -> Tuple[str, Any]:
        """
        LDAPEntry.popitem() -> (k, v), remove and return some (key, value)
        pair as a 2-tuple; but raise KeyError if LDAPEntry is empty.
        """
        try:
            keylist = LDAPValueList(self.keys())
            keylist.remove("dn")
            key = keylist.pop(0)
            value = self[key]
            del self[key]
            return (key, value)
        except IndexError:
            raise KeyError("popitem(): LDAPEntry is empty")

    def __eq__(self, other: Any) -> bool:
        """
        Two LDAPEntry objects are considered equals, if their DN is the same.

        :param other: the other comparable object.
        :return: True if the two object are equals.
        :rtype: bool
        """
        if isinstance(other, self.__class__):
            return self.dn == other.dn
        else:
            return super().__eq__(other)

    def _status(self) -> Dict:
        status = {}
        for key, value in self.items():
            try:
                status[key] = value._status_dict
            except AttributeError:
                continue  # Skip DN
        status["@deleted_keys"] = self.deleted_keys
        return status

    @property
    def extended_dn(self) -> Optional[str]:
        """
        The extended DN of the entry. It is None, if the extended DN control
        is not set or not supported. The attribute is read-only.
        """
        return self.__extended_dn

    @extended_dn.setter
    def extended_dn(self, value: Any) -> None:
        raise ValueError("Extended_dn attribute cannot be set.")

    def change_attribute(self, name: str, optype: int, *values: Tuple) -> None:
        """
        Change an attribute of the entry with explicit LDAP modification type
        by listing the values as parameters.
        An attribute can be removed entirely if the `optype` is delete and no
        `values` are passed.
        This method can be useful for changing write-only attributes (e.g.
        passwords).

        :param str name: the name of the attribute.
        :param int optype: the operation type, 0 for adding, 1 for deleting \
        and 2 for replacing. An :class:`LDAPModOp` also can be used as value.
        :param \*values: the new value or values of the attribute.
        """
        lvl = self.get(name, LDAPValueList())
        if optype == LDAPModOp.ADD:
            lvl.added.extend(values)
            for val in values:
                lvl._append_unchecked(val)
        elif optype == LDAPModOp.DELETE:
            if len(values) == 0:
                self.__setitem__(name, None)
                del self[name]
                return
            else:
                lvl.deleted.extend(values)
                for val in values:
                    lvl._remove_unchecked(val)
        elif optype == LDAPModOp.REPLACE:
            lvl = LDAPValueList(values)
        else:
            raise ValueError("Wrong operation type.")
        self[name] = lvl
        lvl.status = 2 if optype == 2 else 1

    def clear_attribute_changes(self, name: str) -> None:
        """
        Clear all added and deleted changes of an attribute.

        :param str name: the name of the attribute.
        """
        lvl = self.get(name, LDAPValueList())
        lvl.added.clear()
        lvl.deleted.clear()
        self[name] = lvl
        lvl.status = 0

    def items(self, *, exclude_dn: bool = False) -> Union[ItemsView, Iterator]:
        """
        Same functionality as dict.items() but with an optional `exclude_dn`
        keyword-only argument. When `exclude_dn` is True it returns a
        generator of the LDAPEntry's key-value pairs without the dn key and value.

        :param bool exclude_dn: exclude dn key from the list.
        :return: sequence of key-value pairs.
        :rtype: dict_items, generator
        """
        if exclude_dn:
            return (item for item in super().items() if item[0] != "dn")
        else:
            return super().items()

    def keys(self, *, exclude_dn: bool = False) -> Union[KeysView, Iterator]:
        """
        Same functionality as dict.keys() but with an optional `exclude_dn`
        keyword-only argument. When `exclude_dn` is True it returns a
        generator of the LDAPEntry's keys without the dn key.

        :param bool exclude_dn: exclude dn key from the list.
        :return: sequence of keys.
        :rtype: dict_keys, generator
        """
        if exclude_dn:
            return (key for key in super().keys() if key != "dn")
        else:
            return super().keys()

    def values(self, *, exclude_dn: bool = False) -> Union[ValuesView, Iterator]:
        """
        Same functionality as dict.values() but with an optional `exclude_dn`
        keyword-only argument. When `exclude_dn` is True it returns a
        generator of the LDAPEntry's values without the value of the dn key.

        :param bool exclude_dn: exclude dn key from the list.
        :return: sequence of values.
        :rtype: dict_values, generator
        """
        if exclude_dn:
            return (item for item in super().values() if item is not self.dn)
        else:
            return super().values()




import unittest
import sys

import pytest
from conftest import get_config, network_delay

from bonsai import LDAPClient
from bonsai import LDAPEntry
import bonsai.errors


def dummy(timeout=None):
    def dummy_f(f):
        return f

    return dummy_f


try:
    from tornado import gen
    from tornado.testing import gen_test
    from tornado.testing import AsyncTestCase
    from bonsai.tornado import TornadoLDAPConnection

    TestCaseClass = AsyncTestCase
    MOD_INSTALLED = True
except ImportError:
    TestCaseClass = unittest.TestCase
    gen_test = dummy
    MOD_INSTALLED = False


@pytest.mark.skipif(not MOD_INSTALLED, reason="Tornado is not installed.")
class TornadoLDAPConnectionTest(TestCaseClass):
    """ Test TornadoLDAPConnection object. """

    def setUp(self):
        """ Set LDAP URL and open connection. """
        self.cfg = get_config()
        self.url = "ldap://%s:%s/%s?%s?%s" % (
            self.cfg["SERVER"]["hostip"],
            self.cfg["SERVER"]["port"],
            self.cfg["SERVER"]["basedn"],
            self.cfg["SERVER"]["search_attr"],
            self.cfg["SERVER"]["search_scope"],
        )
        self.basedn = self.cfg["SERVER"]["basedn"]
        self.ipaddr = self.cfg["SERVER"]["hostip"]
        self.client = LDAPClient(self.url)
        self.client.set_credentials(
            "SIMPLE",
            user=self.cfg["SIMPLEAUTH"]["user"],
            password=self.cfg["SIMPLEAUTH"]["password"],
        )
        self.client.set_async_connection_class(TornadoLDAPConnection)
        self.io_loop = self.get_new_ioloop()

    @gen_test(timeout=20.0)
    def test_connection(self):
        """ Test opening a connection. """
        conn = yield self.client.connect(True, ioloop=self.io_loop)
        assert conn is not None
        assert not conn.closed
        conn.close()

    @gen_test(timeout=20.0)
    def test_search(self):
        """ Test search. """
        with (yield self.client.connect(True, ioloop=self.io_loop)) as conn:
            res = yield conn.search()
            assert res is not None

    @gen_test(timeout=20.0)
    def test_add_and_delete(self):
        """ Test addding and deleting an LDAP entry. """
        with (yield self.client.connect(True, ioloop=self.io_loop)) as conn:
            entry = LDAPEntry("cn=async_test,%s" % self.basedn)
            entry["objectclass"] = [
                "top",
                "inetOrgPerson",
                "person",
                "organizationalPerson",
            ]
            entry["sn"] = "async_test"
            try:
                yield conn.add(entry)
            except bonsai.errors.AlreadyExists:
                yield conn.delete(entry.dn)
                yield conn.add(entry)
            except:
                self.fail("Unexpected error.")
            res = yield conn.search()
            assert entry in res
            yield entry.delete()
            res = yield conn.search()
            assert entry not in res

    @gen_test(timeout=20.0)
    def test_recursive_delete(self):
        """ Test removing a subtree recursively. """
        org1 = bonsai.LDAPEntry("ou=testusers,%s" % self.basedn)
        org1.update({"objectclass": ["organizationalUnit", "top"], "ou": "testusers"})
        org2 = bonsai.LDAPEntry("ou=tops,ou=testusers,%s" % self.basedn)
        org2.update({"objectclass": ["organizationalUnit", "top"], "ou": "tops"})
        entry = bonsai.LDAPEntry("cn=tester,ou=tops,ou=testusers,%s" % self.basedn)
        entry.update(
            {"objectclass": ["top", "inetorgperson"], "cn": "tester", "sn": "example"}
        )
        try:
            with (
                yield self.client.connect(True, timeout=10.0, ioloop=self.io_loop)
            ) as conn:
                yield conn.add(org1)
                yield conn.add(org2)
                yield conn.add(entry)
                with pytest.raises(bonsai.errors.NotAllowedOnNonleaf):
                    yield conn.delete(org1.dn)
                yield conn.delete(org1.dn, recursive=True)
                res = yield conn.search(org1.dn, 2)
                assert res == []
        except bonsai.LDAPError as err:
            self.fail("Recursive delete is failed: %s" % err)

    @gen_test(timeout=20.0)
    def test_modify_and_rename(self):
        """ Test modifying and renaming an LDAP entry. """
        with (yield self.client.connect(True, ioloop=self.io_loop)) as conn:
            entry = LDAPEntry("cn=async_test,%s" % self.basedn)
            entry["objectclass"] = [
                "top",
                "inetOrgPerson",
                "person",
                "organizationalPerson",
            ]
            entry["sn"] = "async_test"
            oldname = "cn=async_test,%s" % self.basedn
            newname = "cn=async_test2,%s" % self.basedn
            res = yield conn.search(newname, 0)
            if res:
                yield res[0].delete()
            try:
                yield conn.add(entry)
            except bonsai.errors.AlreadyExists:
                yield conn.delete(entry.dn)
                yield conn.add(entry)
            except:
                self.fail("Unexpected error.")
            entry["sn"] = "async_test2"
            yield entry.modify()
            yield entry.rename(newname)
            res = yield conn.search(entry.dn, 0, attrlist=["sn"])
            assert entry["sn"] == res[0]["sn"]
            res = yield conn.search(oldname, 0)
            assert res == []
            yield conn.delete(entry.dn)

    @gen_test(timeout=20.0)
    def test_obj_err(self):
        """ Test object class violation error. """
        entry = LDAPEntry("cn=async_test,%s" % self.basedn)
        entry["cn"] = ["async_test"]
        with (yield self.client.connect(True, ioloop=self.io_loop)) as conn:
            with pytest.raises(bonsai.errors.ObjectClassViolation):
                yield conn.add(entry)

    @gen_test(timeout=20.0)
    def test_whoami(self):
        """ Test whoami. """
        with (yield self.client.connect(True, ioloop=self.io_loop)) as conn:
            obj = yield conn.whoami()
            expected_res = [
                "dn:%s" % self.cfg["SIMPLEAUTH"]["user"],
                self.cfg["SIMPLEAUTH"]["adusername"],
            ]
            assert obj in expected_res

    @gen_test(timeout=12.0)
    def test_connection_timeout(self):
        """ Test connection timeout. """
        with network_delay(7.0):
            with pytest.raises(gen.TimeoutError):
                yield self.client.connect(True, ioloop=self.io_loop, timeout=8.0)

    @gen_test(timeout=18.0)
    def test_search_timeout(self):
        """ Test search timeout. """
        with (yield self.client.connect(True, ioloop=self.io_loop)) as conn:
            with network_delay(7.0):
                with pytest.raises(gen.TimeoutError):
                    yield conn.search(timeout=3.0)

    @pytest.mark.skipif(
        sys.version_info.minor < 5,
        reason="No __aiter__ and __anext__ methods under 3.5.",
    )
    @gen_test(timeout=20.0)
    def test_paged_search(self):
        """ Test paged search. """
        search_dn = "ou=nerdherd,%s" % self.basedn
        with (yield self.client.connect(True, ioloop=self.io_loop)) as conn:
            # To keep compatibility with 3.4 it does not uses async for,
            # but its while loop equivalent.
            res_iter = yield conn.paged_search(search_dn, 1, page_size=3)
            res_iter = type(res_iter).__aiter__(res_iter)
            cnt = 0
            while True:
                try:
                    res = yield type(res_iter).__anext__(res_iter)
                    assert isinstance(res, LDAPEntry)
                    cnt += 1
                except StopAsyncIteration:
                    break
            assert cnt == 6



from typing import List, Any, Union

from .ldapclient import LDAPClient
from .ldapurl import LDAPURL
from .errors import LDAPError


class LDAPReference:
    """
    Object for handling an LDAP reference.

    :param LDAPClient client: a client object.
    :param list references: list of valid LDAP URLs (as string or \
    :class:`LDAPURL` objects).
    """

    def __init__(
        self, client: LDAPClient, references: List[Union[str, LDAPURL]]
    ) -> None:
        if type(client) != LDAPClient:
            raise TypeError("Client parameter must be an LDAPClient.")
        self.__client = client
        self.__refs = []  # type: List[LDAPURL]
        for ref in references:
            if isinstance(ref, str):
                self.__refs.append(LDAPURL(ref))
            elif isinstance(ref, LDAPURL):
                self.__refs.append(ref)
            else:
                raise TypeError("Reference must be string or LDAPURL.")

    @property
    def client(self) -> LDAPClient:
        """ The LDAP client. """
        return self.__client

    @client.setter
    def client(self, value: LDAPClient) -> None:
        if type(value) != LDAPClient:
            raise TypeError("Client property must be an LDAPClient.")
        self.__client = value

    @property
    def references(self) -> List[LDAPURL]:
        """ The list of LDAPURLs of the references. """
        return self.__refs

    @references.setter
    def references(self, value: Any) -> None:
        raise ValueError("The references attribute cannot be set.")



import sys
import bonsai.errors
from bonsai import get_vendor_info
from bonsai import LDAPClient
from bonsai import LDAPEntry

import pytest
from conftest import get_config, network_delay

try:
    from gevent import socket
    from bonsai.gevent import GeventLDAPConnection
except ImportError:
    pass
gevent = pytest.importorskip("gevent")


@pytest.fixture(scope="module")
def gclient():
    """ Get an LDAPClient with GeventLDAPConnection async class. """
    cfg = get_config()
    url = "ldap://%s:%s/%s?%s?%s" % (
        cfg["SERVER"]["hostip"],
        cfg["SERVER"]["port"],
        cfg["SERVER"]["basedn"],
        cfg["SERVER"]["search_attr"],
        cfg["SERVER"]["search_scope"],
    )
    cli = LDAPClient(url)
    cli.set_credentials(
        "SIMPLE", user=cfg["SIMPLEAUTH"]["user"], password=cfg["SIMPLEAUTH"]["password"]
    )
    cli.set_async_connection_class(GeventLDAPConnection)
    return cli


def test_connection(gclient):
    """ Test opening a connection. """
    conn = gclient.connect(True)
    assert conn is not None
    assert not conn.closed


def test_search(gclient):
    """ Test search. """
    with gclient.connect(True) as conn:
        res = conn.search()
        assert res is not None


def test_add_and_delete(gclient, basedn):
    """ Test adding and deleting an LDAP entry. """
    with gclient.connect(True) as conn:
        entry = LDAPEntry("cn=async_test,%s" % basedn)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        entry["sn"] = "async_test"
        try:
            conn.add(entry)
        except bonsai.errors.AlreadyExists:
            conn.delete(entry.dn)
            conn.add(entry)
        except:
            pytest.fail("Unexpected error.")
        res = conn.search()
        assert entry in res
        entry.delete()
        res = conn.search()
        assert entry not in res


def test_recursive_delete(gclient, basedn):
    """ Test removing a subtree recursively. """
    org1 = bonsai.LDAPEntry("ou=testusers,%s" % basedn)
    org1.update({"objectclass": ["organizationalUnit", "top"], "ou": "testusers"})
    org2 = bonsai.LDAPEntry("ou=tops,ou=testusers,%s" % basedn)
    org2.update({"objectclass": ["organizationalUnit", "top"], "ou": "tops"})
    entry = bonsai.LDAPEntry("cn=tester,ou=tops,ou=testusers,%s" % basedn)
    entry.update(
        {"objectclass": ["top", "inetorgperson"], "cn": "tester", "sn": "example"}
    )
    try:
        with gclient.connect(True) as conn:
            conn.add(org1)
            conn.add(org2)
            conn.add(entry)
            with pytest.raises(bonsai.errors.NotAllowedOnNonleaf):
                conn.delete(org1.dn)
            conn.delete(org1.dn, recursive=True)
            res = conn.search(org1.dn, 2)
            assert res == []
    except bonsai.LDAPError as err:
        pytest.fail("Recursive delete is failed: %s" % err)


def test_modify_and_rename(gclient, basedn):
    """ Test modifying and renaming LDAP entry. """
    with gclient.connect(True) as conn:
        entry = LDAPEntry("cn=async_test,%s" % basedn)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        entry["sn"] = "async_test"
        oldname = "cn=async_test,%s" % basedn
        newname = "cn=async_test2,%s" % basedn
        res = conn.search(newname, 0)
        if res:
            res[0].delete()
        try:
            conn.add(entry)
        except bonsai.errors.AlreadyExists:
            conn.delete(entry.dn)
            conn.add(entry)
        except:
            pytest.fail("Unexpected error.")
        entry["sn"] = "async_test2"
        entry.modify()
        entry.rename(newname)
        res = conn.search(entry.dn, 0, attrlist=["sn"])
        assert entry["sn"] == res[0]["sn"]
        res = conn.search(oldname, 0)
        assert res == []
        conn.delete(entry.dn)


def test_obj_err(gclient, basedn):
    """ Test object class violation error. """
    entry = LDAPEntry("cn=async_test,%s" % basedn)
    entry["cn"] = ["async_test"]
    with pytest.raises(bonsai.errors.ObjectClassViolation):
        with gclient.connect(True) as conn:
            conn.add(entry)


def test_whoami(gclient):
    """ Test whoami. """
    with gclient.connect(True) as conn:
        cfg = get_config()
        obj = conn.whoami()
        expected_res = [
            "dn:%s" % cfg["SIMPLEAUTH"]["user"],
            cfg["SIMPLEAUTH"]["adusername"],
        ]
        assert obj in expected_res


@pytest.mark.skipif(
    get_vendor_info()[1] < 20445 or sys.platform != "linux",
    reason="No async timeout support",
)
@pytest.mark.timeout(18)
def test_connection_timeout(gclient):
    """ Test connection timeout. """
    with network_delay(6.0):
        with pytest.raises(socket.timeout):
            gclient.connect(True, timeout=5.0)



"""
.. module:: LDAPClient
   :platform: Unix, Windows
   :synopsis: For managing LDAP connections.

"""
import socket
import sys
from typing import Any, Union, List, Tuple, Optional, Dict, TypeVar

from .ldapurl import LDAPURL
from .ldapconnection import BaseLDAPConnection, LDAPConnection
from .ldapconnection import LDAPSearchScope
from .ldapentry import LDAPEntry

# XXX: The following Python 3.4 guard should be removed in next release.
if sys.version_info >= (3, 5):
    from .asyncio import AIOLDAPConnection
else:
    import warnings
    warnings.warn(
        "Python 3.4 (and earlier) is unsupported, asyncio based AIOLDAPConnection cannot be used.",
        DeprecationWarning
    )
    AIOLDAPConnection = None

CT = TypeVar("CT", bound=BaseLDAPConnection)


class LDAPClient:
    """
    A class for configuring the connection to the directory server.

    :param str|LDAPURL url: an LDAP URL.
    :param bool tls: Set `True` to use TLS connection.
    :raises TypeError: if the `url` parameter is not string \
    or not a valid LDAP URL.
    """

    def __init__(self, url: Union[LDAPURL, str] = "ldap://", tls: bool = False) -> None:
        """ init method. """
        self.__tls = tls
        self.set_url(url)
        self.__credentials = None  # type: Optional[Dict[str, Optional[str]]]
        self.__raw_list = []  # type: List[str]
        self.__mechanism = "SIMPLE"
        self.__cert_policy = -1
        self.__ca_cert = ""  # type: Optional[str]
        self.__ca_cert_dir = ""  # type: Optional[str]
        self.__client_cert = ""  # type: Optional[str]
        self.__client_key = ""  # type: Optional[str]
        self.__async_conn = AIOLDAPConnection
        self.__ppolicy_ctrl = False
        self.__ext_dn = None  # type: Optional[int]
        self.__auto_acquire = True
        self.__chase_referrals = True
        self.__managedsait_ctrl = False

    @staticmethod
    def _create_socketpair() -> Tuple[socket.socket, socket.socket]:
        """
        Create a socketpair that will be used for signaling to select() during
        the initialisation procedure (and binding on MS Windows).
        """
        if hasattr(socket, "socketpair"):
            return socket.socketpair()
        # Backward compatibility on Windows from Python 3.5.
        # Origin: https://gist.github.com/4325783, by Geert Jansen. Public domain.
        def socketpair(
            family: Any = socket.AF_INET, type: Any = socket.SOCK_STREAM, proto: int = 0
        ) -> Tuple[socket.socket, socket.socket]:
            import errno

            # We create a connected TCP socket. Note the trick with setblocking(0)
            # that prevents us from having to create a thread.
            lsock = socket.socket(family, type, proto)
            lsock.bind(("localhost", 0))
            lsock.listen(1)
            addr, port = lsock.getsockname()
            csock = socket.socket(family, type, proto)
            csock.setblocking(0)
            try:
                csock.connect((addr, port))
            except socket.error as serr:
                if serr.errno != errno.WSAEWOULDBLOCK:
                    raise
            ssock, addr = lsock.accept()
            csock.setblocking(1)
            lsock.close()
            return (ssock, csock)

        return socketpair()

    def set_raw_attributes(self, raw_list: List[str]) -> None:
        """
        By default the values of the LDAPEntry are in string format. The
        values of the listed LDAP attribute's names in `raw_list` will be
        kept in bytearray format.

        :param list raw_list: a list of LDAP attribute's names. \
        The elements must be string and unique.

        :raises TypeError: if any of the list's element is not a \
        string.
        :raises ValueError: if the item in the lit is not a unique \
        element.
        """
        for elem in raw_list:
            if not isinstance(elem, str):
                raise TypeError("All element of raw_list must be string.")
        if len(raw_list) > len(set(map(str.lower, raw_list))):
            raise ValueError("Attribute names must be different from each other.")
        self.__raw_list = raw_list

    def set_credentials(
        self,
        mechanism: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        realm: Optional[str] = None,
        authz_id: Optional[str] = None,
        keytab: Optional[str] = None,
    ) -> None:
        """
        Set binding mechanism and credential information. All parameters \
        are optional except the `mechanism`. Different mechanism applies \
        different credential information and ignores the rest. For example:

            * *SIMPLE* uses the `user` (as bind DN) and `password`.
            * *EXTERNAL* only uses the `authz_id` as authorization ID.

        For other use-cases see this section about \
        :ref:`authentication mechanisms <auth-mechs>`.

        :param str mechanism: the name of the binding mechanism.
        :param str user: the identification of the binding user.
        :param str password: the password of the user.
        :param str realm: the (Kerberos) realm of the user.
        :param str authz_id: the authorization ID for the user.
        :param str keytab: path to a Kerberos keytab for authentication.
        :raises TypeError: if mechanism is not string, or any of the other \
        parameters are not string or None.
        """
        if not isinstance(mechanism, str):
            raise TypeError("The mechanism must be a string.")
        mechanism = mechanism.upper()
        creds = {
            "user": user,
            "password": password,
            "realm": realm,
            "authz_id": authz_id,
            "keytab": keytab,
        }
        if (
            list(filter(lambda x: not isinstance(x, (str, type(None))), creds.values()))
            != []
        ):
            raise TypeError("Every parameter must be a string or None.")
        self.__mechanism = mechanism.upper()
        self.__credentials = creds

    def set_cert_policy(self, policy: str) -> None:
        """
        Set policy about server certification.

        :param str policy: the cert policy could be one of the following \
        strings:

            - `try` or `demand`: the server cert will be verified, and if it \
            fail, then the :meth:`LDAPClient.connect` will raise an error.
            - `never` or `allow`: the server cert will be used without any \
            verification.

        :raises TypeError: if the `policy` parameter is not a string.
        :raises ValueError: if the `policy` not one of the four above.

        .. warning::
           Set off the cert verification is dangerous. Without verification \
           there is a chance of man-in-the-middle attack.
        """
        tls_options = {"never": 0, "demand": 2, "allow": 3, "try": 4}
        if not isinstance(policy, str):
            raise TypeError("Policy parameter must be string.")
        policy = policy.lower()
        if policy not in tls_options.keys():
            raise ValueError("'%s' is an invalid policy.", policy)
        self.__cert_policy = tls_options[policy]

    def set_ca_cert(self, name: Optional[str]) -> None:
        """
        Set the name of CA certificate. If the underlying libldap library \
        uses the Mozilla NSS as TLS library the `name` should be the same \
        one in the cert/key database (that specified with \
        :meth:`LDAPClient.set_ca_cert_dir`), otherwise it can be the name \
        of the CA cert file.

        .. note::
           This method has no effect on MS Windows, because WinLDAP \
           searches for the corresponding CA certificate in the cert \
           store. This means that the necessary certificates have to be \
           installed manually in to the cert store.

        :param str name: the name of the CA cert.
        :raises TypeError: if `name` parameter is not a string or not None.
        """
        if name is not None and not isinstance(name, str):
            raise TypeError("Name parameter must be string or None.")
        self.__ca_cert = name

    def set_ca_cert_dir(self, path: Optional[str]) -> None:
        """
        Set the directory of the CA cert. If the underlying libldap library \
        uses the Mozilla NSS as TLS library the `path` should be the path to \
        the existing cert/key database, otherwise it can be the path of the \
        CA cert file.

        .. note::
           This method has no effect on MS Windows, because WinLDAP \
           searches for the corresponding CA certificate in the cert \
           store. This means that the necessary certifications have to be \
           installed manually in to the cert store.

        :param str path: the path to the CA directory.
        :raises TypeError: if `path` parameter is not a string or not None.
        """
        if path is not None and not isinstance(path, str):
            raise TypeError("Path parameter must be string or None.")
        self.__ca_cert_dir = path

    def set_client_cert(self, name: Optional[str]) -> None:
        """
        Set the name of client certificate. If the underlying libldap library \
        uses the Mozilla NSS as TLS library the `name` should be the same one \
        in the cert/key database (that specified with \
        :meth:`LDAPClient.set_ca_cert_dir`), otherwise it can be the name \
        of the client certificate file.

        .. note::
           This method has no effect on MS Windows, because WinLDAP \
           searches for the corresponding client certificate based on \
           the servert's CA cert in the cert store. This means that the \
           necessary certificates have to be installed manually in to \
           the cert store.

        :param str name: the name of the client cert.
        :raises TypeError: if `name` parameter is not a string or not None.
        """
        if name is not None and not isinstance(name, str):
            raise TypeError("Name parameter must be string or None.")
        self.__client_cert = name

    def set_client_key(self, name: Optional[str]) -> None:
        """
        Set the file that contains the private key that matches the \
        certificate of the client that specified with \
        :meth:`LDAPClient.set_client_cert`).

        .. note::
           This method has no effect on MS Windows, because WinLDAP \
           searches for the corresponding client certificate based on \
           the servert's CA cert in the cert store. This means that the \
           necessary certificates have to be installed manually in to \
           the cert store.

        :param str name: the name of the CA cert.
        :raises TypeError: if `name` parameter is not a string or not None.
        """
        if name is not None and not isinstance(name, str):
            raise TypeError("Name parameter must be string or None.")
        self.__client_key = name

    def set_async_connection_class(self, conn: BaseLDAPConnection) -> None:
        """
        Set the LDAP connection class for asynchronous connection. The \
        default connection class is :class:`bonsai.asyncio.AIOLDAPConnection`
        that uses the asyncio event loop.

        :param BaseLDAPConnection conn: the new asynchronous connection class \
        that is a subclass of LDAPConnection.
        :raises TypeError: if `conn` parameter is not a subclass \
        of :class:`BaseLDAPConnection`.
        """
        if not issubclass(conn, BaseLDAPConnection):
            raise TypeError("Class must be a subclass of BaseLDAPConnection. ")
        self.__async_conn = conn

    def set_password_policy(self, ppolicy: bool) -> None:
        """
        Enable password policy control, if it is provided by the directory \
        server. Setting it `True` will change the return value of \
        :meth:`LDAPClient.connect` and :meth:`LDAPConnection.open` to a \
        tuple of `(conn, ctrl)` where the `conn` is an \
        :class:`LDAPConnection`, the `ctrl` is a dict of returned password \
        policy control response that contains the oid, the remaining seconds \
        of password expiration, and the number of remaining grace logins. \
        If the password policy control is not available on the server or not \
        supported by the platform the second item in the returned tuple is \
        `None`, instead of a dictionary.

        By enabling the password policy control the server can send \
        additional error messages related to the user's account and \
        password during conneting to the server and changing entries.

        :param bool ppolicy: enabling/disabling password policy control.
        :raises TypeError: If the parameter is not a bool type.
        """
        if not isinstance(ppolicy, bool):
            raise TypeError("Parameter must be bool.")
        self.__ppolicy_ctrl = ppolicy

    def set_extended_dn(self, extdn_format: Optional[int]) -> None:
        """
        Set the format of extended distinguished name for \
        LDAP_SERVER_EXTENDED_DN_OID control which extends the entries'
        distingushed name with GUID and SID attributes. If the server
        supports the control, the LDAPEntry objects' `extended_dn` attribute
        will be set (as a string) and the `dn` attribute will be kept in
        the simple format.

        Setting 0 specifies that the GUID and SID values be returned in \
        hexadecimal string format, while setting 1 will return the GUID and \
        SID values in standard string format. Passing `None` will remove the \
        control in a format of `<GUID=xxxx>;<SID=yyyy>;distinguishedName`.

        :param int extdn_format: the format of the extended dn. It can be 0, \
        1 or `None`.
        :raises TypeError: if the parameter is not int or None.
        :raises ValueError: if the parameter is not 0, 1 or None.
        """
        if extdn_format is not None and not isinstance(extdn_format, int):
            raise TypeError("Parameter's type must be int or None.")
        if extdn_format not in (0, 1, None):
            raise ValueError("Parameter must be 0, 1 or None.")
        self.__ext_dn = extdn_format

    def set_auto_page_acquire(self, val: bool) -> None:
        """
        Turn on or off the automatic page acquiring during a paged
        LDAP search. By turning automatic page acquiring on, it is
        unnecessary to call :meth:`ldapsearchiter.acquire_next_page`.
        It will be implicitly called during iteration.

        :param bool val: enabling/disabling auto page acquiring.
        :raises TypeError: If the parameter is not a bool type.
        """
        if not isinstance(val, bool):
            raise TypeError("Parameter's type must be bool.")
        self.__auto_acquire = val

    def set_server_chase_referrals(self, val: bool) -> None:
        """
        Turn on or off chasing LDAP referrals by the server. By turning
        off server-side referral chasing search result can contain
        :class:`LDAPReference` objects along with :class:`LDAPEntry`
        objects.

        :param bool val: enabling/disabling LDAP referrals chasing.
        :raises TypeError: If the parameter is not a bool type.
        """
        if not isinstance(val, bool):
            raise TypeError("Parameter's type must be bool.")
        self.__chase_referrals = val

    def set_managedsait(self, val: bool) -> None:
        """
        Set ManageDsaIT control for LDAP operations. With ManageDsaIT an
        LDAP referral can be searched, added and modified as a common
        LDAP entry.

        :param bool val: enabling/disabling ManageDsaIT control.
        :raises TypeError: If the parameter is not a bool type.
        """
        if not isinstance(val, bool):
            raise TypeError("Parameter's type must be bool.")
        self.__managedsait_ctrl = val

    def set_url(self, url: Union[LDAPURL, str]) -> None:
        """
        Set LDAP url for the client.

        :param LDAPURL|str url: the LDAP url.
        """
        if isinstance(url, str):
            self.__url = LDAPURL(url)
        elif isinstance(url, LDAPURL):
            self.__url = url
        else:
            raise TypeError("The url parameter must be string or an LDAPURL.")
        if self.__url.scheme != "ldaps" and self.__tls:
            self.__tls = True
        else:
            self.__tls = False

    @property
    def url(self) -> LDAPURL:
        """ The URL of the directory server. """
        return self.__url

    @url.setter
    def url(self, value: Union[LDAPURL, str]) -> None:
        self.set_url(value)

    @property
    def mechanism(self) -> str:
        """ The chosen mechanism for authentication. It cannot be set. """
        return self.__mechanism

    @mechanism.setter
    def mechanism(self, value: Any = None) -> None:
        raise ValueError("Mechanism attribute cannot be set.")

    @property
    def credentials(self) -> Optional[Dict[str, Optional[str]]]:
        """ A dict with the credential information. It cannot be set. """
        return self.__credentials

    @credentials.setter
    def credentials(self, value: Any = None) -> None:
        raise ValueError("Credentials attribute cannot be set.")

    @property
    def tls(self) -> bool:
        """ A bool about TLS connection is required. It cannot be set."""
        return self.__tls

    @tls.setter
    def tls(self, value: Any = None) -> None:
        raise ValueError("TlS attribute cannot be set.")

    @property
    def cert_policy(self) -> int:
        """ The certification policy. """
        return self.__cert_policy

    @cert_policy.setter
    def cert_policy(self, value: str) -> None:
        self.set_cert_policy(value)

    @property
    def ca_cert(self) -> Optional[str]:
        """ The name of the CA certificate. """
        return self.__ca_cert

    @ca_cert.setter
    def ca_cert(self, value: Optional[str] = None) -> None:
        self.set_ca_cert(value)

    @property
    def ca_cert_dir(self) -> Optional[str]:
        """ The path to the CA certificate. """
        return self.__ca_cert_dir

    @ca_cert_dir.setter
    def ca_cert_dir(self, value: Optional[str]) -> None:
        self.set_ca_cert_dir(value)

    @property
    def client_cert(self) -> Optional[str]:
        """ The name of the client certificate. """
        return self.__client_cert

    @client_cert.setter
    def client_cert(self, value: Optional[str]) -> None:
        self.set_client_cert(value)

    @property
    def client_key(self) -> Optional[str]:
        """ The key file to the client's certificate. """
        return self.__client_key

    @client_key.setter
    def client_key(self, value: Optional[str]) -> None:
        self.set_client_key(value)

    @property
    def raw_attributes(self) -> List[str]:
        """ A list of attributes that should be kept in byte format. """
        return self.__raw_list

    @raw_attributes.setter
    def raw_attributes(self, value: List[str]) -> None:
        self.set_raw_attributes(value)

    @property
    def password_policy(self) -> bool:
        """ The status of using password policy. """
        return self.__ppolicy_ctrl

    @password_policy.setter
    def password_policy(self, value: bool) -> None:
        self.set_password_policy(value)

    @property
    def extended_dn_format(self) -> Optional[int]:
        """
        Format of the extended distinguished name. 0 means hexadecimal string
        format, 1 standard string format. If it is `None`, then it's not set.
        """
        return self.__ext_dn

    @extended_dn_format.setter
    def extended_dn_format(self, value: Optional[int]) -> None:
        self.set_extended_dn(value)

    @property
    def auto_page_acquire(self) -> bool:
        """ The status of automatic page acquiring. """
        return self.__auto_acquire

    @auto_page_acquire.setter
    def auto_page_acquire(self, value: bool) -> None:
        self.set_auto_page_acquire(value)

    @property
    def server_chase_referrals(self) -> bool:
        """ The status of chasing referrals by the server. """
        return self.__chase_referrals

    @server_chase_referrals.setter
    def server_chase_referrals(self, value: bool) -> None:
        self.set_server_chase_referrals(value)

    @property
    def managedsait(self) -> bool:
        """ The status of using ManageDsaIT control. """
        return self.__managedsait_ctrl

    @managedsait.setter
    def managedsait(self, value: bool) -> None:
        self.set_managedsait(value)

    def get_rootDSE(self) -> Optional[LDAPEntry]:
        """
        Returns the server's root DSE entry. The root DSE may contain
        information about the vendor, the naming contexts, the request
        controls the server supports, the supported SASL mechanisms,
        features, schema location, and other information.

        :return: the root DSE entry.
        :rtype: :class:`LDAPEntry`
        """
        attrs = [
            "namingContexts",
            "altServer",
            "supportedExtension",
            "supportedControl",
            "supportedSASLMechanisms",
            "supportedLDAPVersion",
        ]
        tls_options = {0: "never", 2: "demand", 3: "allow", 4: "try", -1: None}
        this = self.__class__(self.url, self.tls)
        cert_policy = tls_options[self.cert_policy]
        if cert_policy is not None:
            this.cert_policy = cert_policy
        this.ca_cert = self.ca_cert
        this.ca_cert_dir = self.ca_cert_dir
        this.client_cert = self.client_cert
        this.client_key = self.client_key
        with LDAPConnection(this).open() as conn:
            try:
                root_dse = conn.search(
                    "", LDAPSearchScope.BASE, "(objectclass=*)", attrs, None, False
                )[0]
                return root_dse
            except IndexError:
                return None

    def connect(
        self,
        is_async: bool = False,
        timeout: Optional[float] = None,
        **kwargs: Dict[str, Any]
    ) -> CT:
        """
        Open a connection to the LDAP server.

        :param bool is_async: Set `True` to use asynchronous connection.
        :param float timeout: time limit in seconds for the operation.
        :param \*\*kwargs: additional keyword arguments that are passed to
                         the async connection object (e.g. an eventloop
                         object as `loop` parameter).
        :return: an LDAP connection.
        :rtype: :class:`LDAPConnection`
        """
        if is_async:
            return self.__async_conn(self, **kwargs).open(timeout)
        else:
            return LDAPConnection(self).open(timeout)




import pytest

from bonsai import LDAPClient
from bonsai.pool import (
    ClosedPool,
    ConnectionPool,
    EmptyPool,
    PoolError,
    ThreadedConnectionPool,
)

import math
import threading
import time


def test_init():
    """ Test pool initialisation. """
    cli = LDAPClient("ldap://dummy.nfo")
    with pytest.raises(ValueError):
        _ = ConnectionPool(cli, minconn=-3)
    with pytest.raises(ValueError):
        _ = ConnectionPool(cli, minconn=5, maxconn=3)
    pool = ConnectionPool(cli, minconn=2, maxconn=5)
    assert pool.closed == True
    assert pool.empty == False
    assert pool.max_connection == 5
    assert pool.shared_connection == 0
    assert pool.idle_connection == 0


def test_open(client):
    """ Test opening the pool. """
    pool = ConnectionPool(client, minconn=5)
    pool.open()
    assert pool.closed != True
    assert pool.idle_connection == 5


def test_close(client):
    """ Test closing the connection. """
    pool = ConnectionPool(client, minconn=1)
    pool.open()
    conn = pool.get()
    pool.close()
    assert pool.closed == True
    assert pool.idle_connection == 0
    assert pool.shared_connection == 0
    assert conn.closed == True


def test_get(client):
    """ Test getting a connection from the pool. """
    pool = ConnectionPool(client, minconn=1, maxconn=2)
    with pytest.raises(ClosedPool):
        _ = pool.get()
    pool.open()
    assert pool.max_connection == 2
    assert pool.idle_connection == 1
    assert pool.shared_connection == 0
    conn1 = pool.get()
    assert conn1 is not None
    assert conn1.closed == False
    assert pool.idle_connection == 0
    assert pool.shared_connection == 1
    conn2 = pool.get()
    assert pool.idle_connection == 0
    assert pool.shared_connection == 2
    with pytest.raises(EmptyPool):
        _ = pool.get()
    assert pool.empty == True


def test_put(client):
    """ Test putting connection back into the pool. """
    pool = ConnectionPool(client, minconn=1, maxconn=1)
    with pytest.raises(ClosedPool):
        pool.put(None)
    pool.open()
    conn = pool.get()
    assert pool.idle_connection == 0
    pool.put(conn)
    assert pool.idle_connection == 1
    other_conn = client.connect()
    with pytest.raises(PoolError):
        pool.put(other_conn)


def test_max_connection():
    """ Test max_connection property. """
    cli = LDAPClient("ldap://dummy.nfo")
    pool = ConnectionPool(cli, minconn=5, maxconn=5)
    assert pool.max_connection == 5
    with pytest.raises(ValueError):
        pool.max_connection = 4
    pool.max_connection = 10
    assert pool.max_connection == 10


def test_spawn(client):
    """ Test context manager. """
    pool = ConnectionPool(client, minconn=1, maxconn=1)
    assert pool.idle_connection == 0
    with pool.spawn() as conn:
        _ = conn.whoami()
    assert pool.idle_connection == 1
    assert pool.shared_connection == 0


def keep(pool, sleep):
    try:
        with pool.spawn() as conn:
            time.sleep(sleep)
            assert not conn.closed
    except ClosedPool:
        pass


def test_threaded_pool_block(client):
    """ Test threaded pool blocks when it's empty. """
    sleep = 5
    pool = ThreadedConnectionPool(client, minconn=1, maxconn=1)
    pool.open()
    t0 = threading.Thread(target=keep, args=(pool, sleep))
    conn = pool.get()
    t0.start()
    pool.put(conn)
    time.sleep(2)
    start = time.time()
    conn = pool.get()
    # Some assertation that it didn't happen immediately.
    assert math.ceil(time.time() - start) + 1.2 >= sleep / 2.0
    assert not conn.closed
    pool.put(conn)


def test_threaded_pool_raise(client):
    """ Test threaded pool blocks when it's empty. """
    sleep = 5
    pool = ThreadedConnectionPool(client, minconn=1, maxconn=1, block=False)
    t0 = threading.Thread(target=keep, args=(pool, sleep))
    t0.start()
    time.sleep(2)
    with pytest.raises(EmptyPool):
        _ = pool.get()
    time.sleep(sleep + 1.2)
    conn = pool.get()
    assert conn.closed == False


def test_threaded_pool_close(client):
    """ Test closing threaded pool. """
    sleep = 5
    pool = ThreadedConnectionPool(client, minconn=1, maxconn=1, block=False)
    t0 = threading.Thread(target=keep, args=(pool, sleep))
    t0.start()
    time.sleep(2)
    pool.close()
    assert pool.closed
    t0.join()


import configparser
import os.path
import sys

import pytest

import bonsai
from bonsai import LDAPReference
from bonsai import LDAPClient
from bonsai import LDAPURL


@pytest.fixture(scope="module")
def host_url():
    """ Set host url for connection. """
    curdir = os.path.abspath(os.path.dirname(__file__))
    cfg = configparser.ConfigParser()
    cfg.read(os.path.join(curdir, "test.ini"))
    url = "ldap://%s:%s" % (cfg["SERVER"]["hostip"], cfg["SERVER"]["port"])
    return url


def test_init_errors(host_url):
    """ Testing errors during initialization of LDAPReference. """
    client = LDAPClient(host_url)
    with pytest.raises(TypeError):
        _ = LDAPReference(None, ["ldap://a"])
    with pytest.raises(TypeError):
        _ = LDAPReference(client, [0])
    with pytest.raises(ValueError):
        _ = LDAPReference(client, ["asd", LDAPURL("ldap://b")])


def test_client_prop(host_url):
    """ Testing client property. """
    client = LDAPClient(host_url)
    ref = LDAPReference(client, [])
    assert ref.client == client
    with pytest.raises(TypeError):
        ref.client = "b"

    ref.client = LDAPClient()
    assert ref.client != client


def test_references_prop(host_url):
    """ Testing references property. """
    client = LDAPClient(host_url)
    reflist = [LDAPURL("ldap://localhost"), host_url]
    ref = LDAPReference(client, reflist)
    assert ref.references == reflist
    with pytest.raises(ValueError):
        ref.references = None


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Referrals are not set in AD."
)
def test_referral_chasing(host_url):
    """ Testing referral chasing option. """
    refdn = "o=admin-ref,ou=nerdherd,dc=bonsai,dc=test"
    client = LDAPClient(host_url)
    with client.connect() as conn:
        res = conn.search(refdn, 0)
        assert isinstance(res[0], bonsai.LDAPEntry)
    client.server_chase_referrals = False
    with client.connect() as conn:
        res = conn.search(refdn, 0)
        assert len(res) == 0
        res = conn.search("ou=nerdherd,dc=bonsai,dc=test", 1)
        refs = [item for item in res if isinstance(item, LDAPReference)]
        assert any(refs)



import base64
import io
import os
from collections import defaultdict
from itertools import groupby
from typing import TextIO, Iterable, Iterator, Any, Optional, List, Union

from .ldapentry import LDAPEntry, LDAPModOp
from .ldapvaluelist import LDAPValueList

from .errors import LDAPError


class LDIFError(LDAPError):
    """ General exception that is raised during reading or writing an LDIF file. """

    code = -300


class LDIFReader:
    """
    Create an object for reading LDAP entries from an LDIF format file \
    as described in RFC 2849.

    :param TextIO input_file: a file-like input object in text mode.
    :param bool autoload: allow to automatically load external \
    sources from URL.
    :param int max_length: the maximal line length of the LDIF file.
    :raises TypeError: if the input_file is not a file-like object \
    or max_length is not an int.
    """

    def __init__(
        self, input_file: TextIO, autoload: bool = True, max_length: int = 76
    ) -> None:
        """ Init method. """
        if not isinstance(max_length, int):
            raise TypeError("The max_length must be int.")
        self.input_file = input_file
        self.autoload = autoload
        self.max_length = max_length
        self.version = None  # type: Optional[int]
        self.__entries = self.__read_attributes()
        self.__num_of_entries = 0
        self.__resource_handlers = {"file": self.__load_file}

    def __read_attributes(self) -> Iterator[List[str]]:
        buffer = []  # type: List[str]
        comment = False
        for num, line in enumerate(self.__file):
            try:
                if len(line) > self.max_length:
                    raise LDIFError("Line {0} is too long.".format(num + 1))
                if len(line.strip()) == 0:
                    yield buffer
                    buffer.clear()
                    continue
                if line[0] == " ":
                    if not comment:
                        # Concat line with the previous one.
                        buffer[-1] = "".join((buffer[-1], line[1:].rstrip()))
                    continue
                elif line[0] == "#":
                    comment = True
                else:
                    comment = False
                if comment:
                    # Drop comment lines.
                    continue
                else:
                    buffer.append(line.rstrip())
            except IndexError:
                raise LDIFError("Parser error at line: {0}.".format(num + 1))
        if buffer:
            yield buffer

    @staticmethod
    def __convert(val: Union[str, bytes]) -> Union[str, bytes, int]:
        try:
            val = int(val)
        except ValueError:
            try:
                val = val.decode("UTF-8")
            except (ValueError, AttributeError):
                pass
        return val

    @staticmethod
    def __find_key(searched_key: str, keylist: List[str]) -> Optional[str]:
        for key in keylist:
            if key.lower() == searched_key.lower():
                return key
        else:
            raise KeyError(searched_key)

    def __load_file(self, url: str) -> bytes:
        _, path = url.split("file://")
        abs_filepath = os.path.normpath(
            os.path.join(os.path.dirname(os.path.abspath(self.__file.name)), path)
        )
        with open(abs_filepath, "rb") as resource:
            return resource.read()

    def load_resource(self, url: str) -> Union[str, bytes]:
        try:
            scheme, _ = url.split(":", maxsplit=1)
            return self.__resource_handlers[scheme](url)
        except (KeyError, ValueError):
            raise LDIFError("Unsupported URL format: {0}.".format(url)) from None

    def __iter__(self) -> "LDIFReader":
        return self

    def __next__(self) -> LDAPEntry:
        entry = LDAPEntry("")
        change_type = "add"
        attr_blocks = [
            list(group)
            for key, group in groupby(next(self.__entries), lambda line: line == "-")
            if not key
        ]
        self.__num_of_entries += 1
        for block in attr_blocks:
            attr_dict = defaultdict(LDAPValueList)
            for attrval in block:
                try:
                    if ":: " in attrval:
                        attr, val = attrval.split(":: ")
                        val = base64.b64decode(val)
                    elif ": " in attrval:
                        attr, val = attrval.split(": ")
                    elif ":< " in attrval:
                        attr, val = attrval.split(":< ")
                        if self.__autoload:
                            val = self.load_resource(val)
                    else:
                        raise ValueError()
                except ValueError:
                    raise LDIFError(
                        "Invalid attribute value pair: '{0}' for entry #{1}.".format(
                            attrval, self.__num_of_entries
                        )
                    ) from None
                if attr.lower() == "changetype":
                    change_type = val.lower()
                elif attr.lower() == "dn":
                    entry.dn = self.__convert(val)
                elif attr.lower() == "version":
                    self.version = self.__convert(val)
                else:
                    attr_dict[attr].append(self.__convert(val))
            if change_type == "modify":
                try:
                    for key in attr_dict.pop("add", []):
                        key = self.__find_key(key, attr_dict.keys())
                        entry.change_attribute(key, LDAPModOp.ADD, *attr_dict[key])
                    for key in attr_dict.pop("replace", []):
                        key = self.__find_key(key, attr_dict.keys())
                        entry.change_attribute(key, LDAPModOp.REPLACE, *attr_dict[key])
                    for key in attr_dict.pop("delete", []):
                        try:
                            key = self.__find_key(key, attr_dict.keys())
                        except KeyError:
                            pass
                        entry.change_attribute(key, LDAPModOp.DELETE, *attr_dict[key])
                except KeyError as err:
                    raise LDIFError(
                        "Missing attribute: '{0}' for entry #{1}.".format(
                            err.args[0], self.__num_of_entries
                        )
                    )
            elif change_type == "add":
                for key, vals in attr_dict.items():
                    entry[key] = vals
        if entry.dn == "":
            raise LDIFError(
                "Missing distinguished name for entry #{0}.".format(
                    self.__num_of_entries
                )
            )
        return entry

    @property
    def input_file(self):
        """ The file-like object of an LDIF file. """
        return self.__file

    @input_file.setter
    def input_file(self, value: io.TextIOBase):
        if not isinstance(value, io.TextIOBase):
            raise TypeError("The input_file must be file-like object in text mode.")
        self.__file = value

    @property
    def autoload(self):
        """ Enable/disable autoloading resources in LDIF files. """
        return self.__autoload

    @autoload.setter
    def autoload(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("The autoload property must be bool.")
        self.__autoload = value

    @property
    def resource_handlers(self):
        """
        A dictionary of supported resource types. The keys are the schemes,
        while the values are functions that expect the full URL parameters
        and return the loaded content in preferably bytes format.
        """
        return self.__resource_handlers

    @resource_handlers.setter
    def resource_handlers(self, value: Any):
        raise ValueError("The resource_handlers attribute cannot be set.")


class LDIFWriter:
    """
    Create an object for serialising LDAP entries in LDIF format as 
    described in RFC 2849.

    :param TextIO output_file: a file-like output object in text mode.
    :param int max_length: the maximal line length of the LDIF file.
    :raises TypeError: if the output_file is not a file-like object \
    or max_length is not an int.
    """

    def __init__(self, output_file: TextIO, max_length: int = 76) -> None:
        """ Init method. """
        if not isinstance(max_length, int):
            raise TypeError("The max_length must be int.")
        self.output_file = output_file
        self.max_length = max_length

    def _get_attr_lines(
        self, attrname: str, attrvalue: Iterable[Any]
    ) -> Iterator[str]:
        for val in attrvalue:
            if isinstance(val, (bytes, bytearray)):
                # If it's a binary has to be base64 encoded anyway.
                has_not_safe_char = True
                has_not_safe_init_char = True
            else:
                val = str(val)
                has_not_safe_char = any(
                    char
                    for char in val
                    if ord(char) > 127 or ord(char) in (0x0A, 0x0D, 0x00)
                ) or val.endswith(" ")
                has_not_safe_init_char = val[0] in (" ", ":", "<")
                val = val.encode("UTF-8")
            if has_not_safe_char or has_not_safe_init_char:
                val = base64.b64encode(val)
                name = "{0}:".format(attrname)  # Add extra colon.
            else:
                name = attrname
            line = "{attr}: {value}".format(attr=name, value=val.decode("UTF-8"))
            for i in range(0, len(line), self.max_length):
                # Split the line into self.max_length.
                if i != 0:
                    yield " {0}\n".format(line[i : i + self.max_length])
                else:
                    yield "{0}\n".format(line[i : i + self.max_length])

    def write_entry(self, entry: LDAPEntry) -> None:
        """
        Write an LDAP entry to the file in LDIF format.

        :param LDAPEntry entry: the LDAP entry to serialise.
        """
        for line in self._get_attr_lines("dn", (entry.dn,)):
            self.__file.write(line)
        for attrname, attrvalue in entry.items(exclude_dn=True):
            for line in self._get_attr_lines(attrname, attrvalue):
                self.__file.write(line)

    def write_entries(
        self, entries: Iterable[LDAPEntry], write_version: bool = True
    ) -> None:
        """
        Write multiple LDAP entry to file in LDIF format, separated
        with newline and with optional version header.

        :param list entries: list of LDAP entries.
        :param bool write_version: if it's True, write version header.
        """
        if write_version:
            self.__file.write(next(self._get_attr_lines("version", (1,))))
        for ent in entries:
            self.write_entry(ent)
            self.__file.write("\n")

    def write_changes(self, entry: LDAPEntry) -> None:
        """
        Write an LDAP entry's changes to file in an LDIF-CHANGE format.
        Only attribute modifications are serialised.

        :param LDAPEntry entry: the LDAP entry to serialise.
        """
        self.__file.write(next(self._get_attr_lines("dn", (entry.dn,))))
        self.__file.write(next(self._get_attr_lines("changetype", ("modify",))))
        changes = dict(entry._status())
        deleted_keys = changes.pop("@deleted_keys")
        for attrname, stat in sorted(changes.items(), key=lambda s: s[1]["@status"]):
            if stat["@status"] == 1 and stat["@added"]:
                self.__file.write(next(self._get_attr_lines("add", (attrname,))))
                for line in self._get_attr_lines(attrname, stat["@added"]):
                    self.__file.write(line)
                self.__file.write("-\n")
            elif stat["@status"] == 1 and stat["@deleted"]:
                self.__file.write(next(self._get_attr_lines("delete", (attrname,))))
                for line in self._get_attr_lines(attrname, stat["@deleted"]):
                    self.__file.write(line)
                self.__file.write("-\n")
            elif stat["@status"] == 2:
                self.__file.write(next(self._get_attr_lines("replace", (attrname,))))
                for line in self._get_attr_lines(attrname, stat["@added"]):
                    self.__file.write(line)
                self.__file.write("-\n")
        for key in deleted_keys:
            self.__file.write(next(self._get_attr_lines("delete", (key,))))
            self.__file.write("-\n")
        self.__file.write("\n")

    @property
    def output_file(self):
        """ The file-like object for an LDIF file. """
        return self.__file

    @output_file.setter
    def output_file(self, value: io.TextIOBase):
        if not isinstance(value, io.TextIOBase):
            raise TypeError("The output_file must be file-like object in text mode.")
        self.__file = value



import configparser
import os
import sys
import tempfile

from contextlib import contextmanager

import distutils.sysconfig
import distutils.ccompiler
from distutils.errors import CompileError, LinkError

try:
    from setuptools import setup, Extension, Command
except ImportError:
    from distutils.core import setup, Extension, Command


@contextmanager
def silent_stderr():
    """ Shush stderr for receiving unnecessary errors during setup. """
    devnull = open(os.devnull, "w")
    old = os.dup(sys.stderr.fileno())
    os.dup2(devnull.fileno(), sys.stderr.fileno())
    try:
        yield devnull
    finally:
        os.dup2(old, sys.stderr.fileno())


def have_krb5(libs, libdirs=None):
    """ Check that the target platform has KRB5 support. """
    code = """
    #include <krb5.h>
    #include <gssapi/gssapi_krb5.h>

    int main(void) {
        unsigned int ms = 0;
        krb5_context ctx;
        const char *cname = NULL;
        gss_key_value_set_desc store;

        store.count = 0;
        krb5_init_context(&ctx);
        gss_krb5_ccache_name(&ms, cname, NULL);
        return 0;
    }
    """
    curdir = os.path.abspath(os.path.dirname(__file__))
    cfg = configparser.ConfigParser()
    cfg.read(os.path.join(curdir, "setup.cfg"))
    with tempfile.TemporaryDirectory() as tmp_dir:
        name = os.path.join(tmp_dir, "test_krb5")
        src_name = name + ".c"
        with open(src_name, "w") as source:
            source.write(code)

        comp = distutils.ccompiler.new_compiler()
        distutils.sysconfig.customize_compiler(comp)
        for include_dir in cfg.get("build_ext", "include_dirs", fallback="").split(":"):
            if include_dir:
                comp.add_include_dir(include_dir)
        for library_dir in cfg.get("build_ext", "library_dirs", fallback="").split(":"):
            if library_dir:
                comp.add_library_dir(library_dir)
        try:
            with silent_stderr():
                if "-coverage" in os.getenv("CFLAGS", ""):
                    # If coverage flag is set.
                    libs.append("gcov")
                comp.link_executable(
                    comp.compile([src_name], output_dir=tmp_dir),
                    name,
                    libraries=libs,
                    library_dirs=libdirs,
                )
        except (CompileError, LinkError):
            return False
        else:
            return True


SOURCES = [
    "bonsaimodule.c",
    "ldapentry.c",
    "ldapconnectiter.c",
    "ldapconnection.c",
    "ldapmodlist.c",
    "ldap-xplat.c",
    "ldapsearchiter.c",
    "utils.c",
]

DEPENDS = [
    "ldapconnection.h",
    "ldapentry.h",
    "ldapconnectiter.h",
    "ldapmodlist.h",
    "ldapsearchiter.h",
    "ldap-xplat.h",
    "utils.h",
]

LIBDIRS = []
MACROS = []
if sys.platform == "darwin":
    LIBDIRS.append("/usr/local/lib")
    MACROS.append(("MACOSX", 1))

if sys.platform == "win32":
    LIBS = ["wldap32", "secur32", "Ws2_32"]
    SOURCES.append("wldap-utf8.c")
    DEPENDS.append("wldap-utf8.h")
    MACROS.append(("WIN32", 1))
else:
    LIBS = ["ldap", "lber"]
    if have_krb5(["krb5", "gssapi"], LIBDIRS):
        LIBS.extend(["krb5", "gssapi"])
        MACROS.append(("HAVE_KRB5", 1))
    elif have_krb5(["krb5", "gssapi_krb5"], LIBDIRS):
        LIBS.extend(["krb5", "gssapi_krb5"])
        MACROS.append(("HAVE_KRB5", 1))
    else:
        print(
            "INFO: Kerberos headers and libraries are not found."
            " Additional GSSAPI capabilities won't be installed."
        )

SOURCES = [os.path.join("src/_bonsai", x) for x in SOURCES]
DEPENDS = [os.path.join("src/_bonsai", x) for x in DEPENDS]

BONSAI_MODULE = Extension(
    "bonsai._bonsai",
    libraries=LIBS,
    sources=SOURCES,
    depends=DEPENDS,
    define_macros=MACROS,
    library_dirs=LIBDIRS,
)
PYTHON_DEPS = []

if sys.version_info.minor < 5:
    # Typing dependecy for Python 3.4 and earlier.
    PYTHON_DEPS.append("typing")

# Get long description from the README.rst file.
with open("README.rst") as file:
    LONG_DESC = file.read()

# Get version number from the module's __init__.py file.
with open("./src/bonsai/__init__.py") as src:
    VER = [
        line.split('"')[1] for line in src.readlines() if line.startswith("__version__")
    ][0]

setup(
    name="bonsai",
    version=VER,
    description="Python 3 module for accessing LDAP directory servers.",
    author="noirello",
    author_email="noirello@gmail.com",
    url="https://github.com/noirello/bonsai",
    long_description=LONG_DESC,
    license="MIT",
    ext_modules=[BONSAI_MODULE],
    package_dir={"bonsai": "src/bonsai"},
    packages=["bonsai", "bonsai.asyncio", "bonsai.gevent", "bonsai.tornado"],
    include_package_data=True,
    install_requires=PYTHON_DEPS,
    extras_require={"gevent": ["gevent>=1.4.0"], "tornado": ["tornado>=5.1.1"]},
    keywords=[
        "python3",
        "ldap",
        "ldap3",
        "python-ldap",
        "libldap",
        "winldap",
        "asyncio",
        "gevent",
        "tornado",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: C",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP",
    ],
)




from ._bonsai import (
    get_tls_impl_name,
    set_connect_async,
    get_vendor_info,
    has_krb5_support,
    _unique_contains,
    set_debug,
)


def escape_attribute_value(attrval: str):
    """
    Escapes the special character in an attribute value
    based on RFC 4514.

    :param str attrval: the attribute value.
    :return: The escaped attribute value.
    :rtype: str
    """
    # Order matters.
    chars_to_escape = ("\\", '"', "+", ",", ";", "<", "=", ">")
    for char in chars_to_escape:
        attrval = attrval.replace(char, "\\{0}".format(char))
    if attrval[0] == "#" or attrval[0] == " ":
        attrval = "".join(("\\", attrval))
    if attrval[-1] == " ":
        attrval = "".join((attrval[:-1], "\\ "))
    attrval = attrval.replace("\0", "\\0")
    return attrval


def escape_filter_exp(filter_exp: str):
    """
    Escapes the special characters in an LDAP filter based on RFC 4515.

    :param str filter_exp: the unescaped filter expression.
    :return: the escaped filter expression.
    :rtype: str
    """
    chars_to_escape = (
        ("\\", "\\5C"),
        ("*", "\\2A"),
        ("(", "\\28"),
        (")", "\\29"),
        ("\0", "\\0"),
    )
    for char, repl in chars_to_escape:
        filter_exp = filter_exp.replace(char, repl)
    return filter_exp



import pytest

from bonsai.ldapvaluelist import LDAPValueList


def test_append():
    """ Test LDAPValueList's append method. """
    lvl = LDAPValueList()
    lvl.append("test")
    assert "test" in lvl
    with pytest.raises(ValueError):
        lvl.append("Test")


def test_insert():
    """ Test LDAPValueList's insert method. """
    lvl = LDAPValueList(("test1",))
    lvl.insert(0, "test2")
    assert lvl == ["test2", "test1"]
    with pytest.raises(ValueError):
        lvl.insert(2, "test2")


def test_remove():
    """ Test LDAPValueList's remove method. """
    lvl = LDAPValueList(("test1", "test2"))
    lvl.remove("Test1")
    assert lvl == ["test2"]
    with pytest.raises(ValueError):
        lvl.remove("test1")


def test_set():
    """ Test LDAPValueList's __setitem__ method. """
    lvl = LDAPValueList()
    lvl[0:2] = ("test1", "test2", "test3")
    lvl[1] = "test4"
    assert lvl == ["test1", "test4", "test3"]
    with pytest.raises(ValueError):
        lvl[1] = "test3"
    with pytest.raises(ValueError):
        lvl[1:3] = ["test5", "test1"]
    del lvl[0:2]
    assert lvl == ["test3"]
    del lvl[0]
    assert lvl == []
    lvl = LDAPValueList([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    del lvl[slice(1, 10, 2)]
    assert lvl == [1, 3, 5, 7, 9, 11, 12]
    lvl[slice(2, 6, 2)] = (13, 14)
    assert lvl == [1, 3, 13, 7, 14, 11, 12]


def test_extend():
    """ Test LDAPValueList's extend method. """
    lvl = LDAPValueList(("test1",))
    lvl.extend(("test2", "test3"))
    assert lvl == ["test1", "test2", "test3"]
    with pytest.raises(ValueError):
        lvl.extend(("test4", "test1"))


def test_pop():
    """ Test LDAPValueList's pop method. """
    lvl = LDAPValueList(("test1", "test2"))
    assert lvl.pop(0) == "test1"
    assert lvl == ["test2"]
    lvl.pop()
    assert lvl == []
    with pytest.raises(IndexError):
        lvl.pop()


def test_copy():
    """ Test LDAPValueList's copy method. """
    lvl1 = LDAPValueList(("test1", "test2"))
    lvl2 = lvl1.copy()
    assert lvl1 == lvl2
    assert lvl1.status == lvl2.status


def test_add():
    """ Test adding list to an LDAPValueList. """
    lvl = LDAPValueList((1, 2, 3))
    assert lvl + [4, 5] == [1, 2, 3, 4, 5]
    with pytest.raises(TypeError):
        _ = lvl + 3
    with pytest.raises(TypeError):
        lvl += "x"
    lvl += [4, 5]
    assert lvl == [1, 2, 3, 4, 5]


def test_mul():
    """ Test multiplying an LDAPValueList. """
    lvl = LDAPValueList((1, 2, 3))
    with pytest.raises(TypeError):
        _ = lvl * 3


def test_set_status():
    """ Test setting LDAPValueList's status. """
    lvl = LDAPValueList()
    with pytest.raises(TypeError):
        lvl.status = "a"
    with pytest.raises(ValueError):
        lvl.status = -1
    lvl.status = 2
    assert lvl.status == 2


def test_clear():
    """ Test setting LDAPValueList's clear method. """
    lvl = LDAPValueList((1, 2, 3))
    lvl.append(4)
    lvl.clear()
    assert lvl == []


def test_readonly_attrs():
    """ Test modifying read-only attributes. """
    lvl = LDAPValueList((1, 2, 3))
    with pytest.raises(ValueError):
        lvl.added = [1, 2, 3]
    with pytest.raises(ValueError):
        lvl.deleted = [1, 2, 3]
    with pytest.raises(TypeError):
        lvl._status_dict = {"status": 2}